package com.minemaarten.signals.block;

import java.util.ArrayList;
import java.util.List;

import net.minecraft.block.properties.PropertyDirection;
import net.minecraft.block.properties.PropertyEnum;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;

import com.minemaarten.signals.api.access.ISignal.EnumLampStatus;
import com.minemaarten.signals.rail.RailManager;
import com.minemaarten.signals.rail.RailWrapper;
import com.minemaarten.signals.tileentity.TileEntitySignalBase;

public class BlockSignalBase extends BlockBase{
    public static PropertyEnum<EnumLampStatus> LAMP_STATUS = PropertyEnum.<EnumLampStatus> func_177709_a("lamp_status", EnumLampStatus.class);
    public static PropertyEnum<EnumFacing> FACING = PropertyDirection.func_177712_a("facing", EnumFacing.Plane.HORIZONTAL);

    public BlockSignalBase(Class<? extends TileEntitySignalBase> tileClass, String name){
        super(tileClass, name);
    }

    @Override
    public AxisAlignedBB func_185496_a(IBlockState state, IBlockAccess source, BlockPos pos){
        return new AxisAlignedBB(4 / 16F, 0.0F, 4 / 16F, 12 / 16F, 8 / 16F, 12 / 16F);
    }

    @Override
    protected BlockStateContainer func_180661_e(){
        return new BlockStateContainer(this, FACING, LAMP_STATUS);
    }

    @Override
    public int func_176201_c(IBlockState state){
        return state.func_177229_b(FACING).func_176736_b() * 4 + state.func_177229_b(LAMP_STATUS).ordinal();
    }

    @Override
    public IBlockState func_176203_a(int meta){
        IBlockState state = super.func_176203_a(meta);
        return state.func_177226_a(FACING, EnumFacing.func_176731_b(meta / 4)).func_177226_a(LAMP_STATUS, EnumLampStatus.values()[meta % 4]);
    }

    @Override
    public IBlockState func_180642_a(World worldIn, BlockPos pos, EnumFacing facing, float hitX, float hitY, float hitZ, int meta, EntityLivingBase placer){
        List<EnumFacing> railFacings = new ArrayList<>(4);
        for(EnumFacing horFacing : EnumFacing.field_176754_o) {
            BlockPos railPos = pos.func_177972_a(horFacing);
            if(RailManager.getInstance().getRail(worldIn, railPos, worldIn.func_180495_p(railPos)) != null) {
                railFacings.add(horFacing);
            }
        }

        EnumFacing orientation;
        if(railFacings.size() == 1) { //When exactly one rail is connecting, connect the signal to the rail
            orientation = railFacings.get(0).func_176746_e();
        } else { //Else, fall back onto player orientation.
            orientation = placer.func_174811_aO();
        }
        return func_176223_P().func_177226_a(FACING, orientation);
    }

    @Override
    public int func_180656_a(IBlockState state, IBlockAccess worldIn, BlockPos pos, EnumFacing side){
        EnumLampStatus lampStatus = state.func_177229_b(LAMP_STATUS);
        if(!(worldIn instanceof WorldServer) || state.func_177230_c() != this || (lampStatus != EnumLampStatus.GREEN && lampStatus != EnumLampStatus.YELLOW) || state.func_177229_b(BlockSignalBase.FACING).func_176746_e() != side) return 0;
        TileEntitySignalBase signal = (TileEntitySignalBase)worldIn.func_175625_s(pos);
        signal.func_145834_a((WorldServer)worldIn);
        for(RailWrapper rail : signal.getConnectedRails()) {
            for(TileEntitySignalBase s : rail.getSignals().values()) {
                if(s != signal) {
                    lampStatus = s.getLampStatus();
                    if(lampStatus != EnumLampStatus.GREEN && lampStatus != EnumLampStatus.YELLOW) return 0;
                }
            }
        }
        return 15;
    }

    /**
     * Can this block provide power. Only wire currently seems to have this change based on its state.
     */
    @Override
    public boolean func_149744_f(IBlockState state){
        return true;
    }

    @Override
    public boolean func_149662_c(IBlockState state){
        return false;
    }

    @Override
    public boolean func_149686_d(IBlockState state){
        return false;
    }

}
