package com.minemaarten.signals.capabilities;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import net.minecraft.block.BlockHopper;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.inventory.InventoryBasic;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.tileentity.TileEntityFurnace;
import net.minecraft.tileentity.TileEntityHopper;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.CapabilityInject;
import net.minecraftforge.common.capabilities.CapabilityManager;
import net.minecraftforge.common.capabilities.ICapabilitySerializable;
import net.minecraftforge.event.entity.minecart.MinecartUpdateEvent;
import net.minecraftforge.fml.common.network.NetworkRegistry;
import net.minecraftforge.items.IItemHandler;
import net.minecraftforge.items.ItemHandlerHelper;
import net.minecraftforge.items.wrapper.InvWrapper;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.Validate;

import com.minemaarten.signals.api.access.IDestinationAccessor;
import com.minemaarten.signals.api.access.ISignal.EnumLampStatus;
import com.minemaarten.signals.init.ModItems;
import com.minemaarten.signals.inventory.EngineItemHandler;
import com.minemaarten.signals.lib.SignalsUtils;
import com.minemaarten.signals.network.GuiSynced;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketSpawnParticle;
import com.minemaarten.signals.network.PacketUpdateMinecartEngineState;
import com.minemaarten.signals.network.PacketUpdateMinecartPath;
import com.minemaarten.signals.rail.DestinationPathFinder.AStarRailNode;
import com.minemaarten.signals.rail.RailCacheManager;
import com.minemaarten.signals.rail.RailWrapper;
import com.minemaarten.signals.tileentity.IGUITextFieldSensitive;
import com.minemaarten.signals.tileentity.TileEntitySignalBase;

public class CapabilityMinecartDestination implements IGUITextFieldSensitive, IDestinationAccessor{
    @CapabilityInject(CapabilityMinecartDestination.class)
    public static Capability<CapabilityMinecartDestination> INSTANCE;
    private static final Pattern EMPTY_PATTERN = Pattern.compile("");

    @GuiSynced
    private String destinationStations = ""; //'\n' separated list of destinations
    private Pattern[] destinationRegexes = new Pattern[0]; //Cache of the regexes of the destinations
    @GuiSynced
    private int curDestinationIndex;
    @GuiSynced
    private String invalidDestinations = ""; //Destinations become invalid when the regex is invalid.

    private AStarRailNode curPath;
    private List<BlockPos> nbtLoadedPath;

    private boolean motorized; //True when an engine upgrade has been applied.
    @GuiSynced
    private int fuelLeft;
    @GuiSynced
    private int totalBurnTime;
    private int hopperTimer;

    private final InventoryBasic fuelInv = new InventoryBasic("cartEngineInv", true, 5){
        @Override
        public boolean func_94041_b(int index, ItemStack stack){
            return stack == null || TileEntityFurnace.func_145954_b(stack);
        }
    };
    private final IItemHandler fuelItemHandler = new InvWrapper(fuelInv);
    private final IItemHandler engineItemHandler = new EngineItemHandler(this, fuelItemHandler);
    private boolean motorActive;
    public boolean travelingBetweenDimensions;

    public static void register(){
        CapabilityManager.INSTANCE.register(CapabilityMinecartDestination.class, new Capability.IStorage<CapabilityMinecartDestination>(){
            @Override
            public NBTBase writeNBT(Capability<CapabilityMinecartDestination> capability, CapabilityMinecartDestination instance, EnumFacing side){
                NBTTagCompound tag = new NBTTagCompound();

                tag.func_74778_a("destinations", instance.destinationStations);
                tag.func_74768_a("destIndex", instance.curDestinationIndex);

                if(instance.curPath != null) {
                    AStarRailNode curNode = instance.curPath;
                    NBTTagList nodeList = new NBTTagList();
                    while(curNode != null) {
                        NBTTagCompound nodeTag = new NBTTagCompound();
                        nodeTag.func_74768_a("x", curNode.getRail().func_177958_n());
                        nodeTag.func_74768_a("y", curNode.getRail().func_177956_o());
                        nodeTag.func_74768_a("z", curNode.getRail().func_177952_p());
                        nodeList.func_74742_a(nodeTag);

                        curNode = curNode.getNextNode();
                    }
                    tag.func_74782_a("path", nodeList);
                }

                tag.func_74757_a("motorized", instance.motorized);
                if(instance.motorized) {
                    tag.func_74768_a("fuelLeft", instance.fuelLeft);
                    tag.func_74768_a("totalBurnTime", instance.totalBurnTime);
                    SignalsUtils.writeInventoryToNBT(tag, instance.fuelInv);
                }

                return tag;
            }

            @Override
            public void readNBT(Capability<CapabilityMinecartDestination> capability, CapabilityMinecartDestination instance, EnumFacing side, NBTBase base){
                NBTTagCompound tag = (NBTTagCompound)base;

                instance.destinationStations = tag.func_74779_i("destinations");
                instance.recompileRegexes();
                instance.curDestinationIndex = tag.func_74762_e("destIndex");

                if(tag.func_74764_b("path")) {
                    instance.nbtLoadedPath = new ArrayList<>();
                    NBTTagList nodeList = tag.func_150295_c("path", 10);
                    for(int i = 0; i < nodeList.func_74745_c(); i++) {
                        NBTTagCompound nodeTag = nodeList.func_150305_b(i);
                        instance.nbtLoadedPath.add(new BlockPos(nodeTag.func_74762_e("x"), nodeTag.func_74762_e("y"), nodeTag.func_74762_e("z")));
                    }
                } else {
                    instance.curPath = null;
                }

                instance.motorized = tag.func_74767_n("motorized");
                if(instance.motorized) {
                    instance.fuelLeft = tag.func_74762_e("fuelLeft");
                    instance.totalBurnTime = tag.func_74762_e("totalBurnTime");
                    SignalsUtils.readInventoryFromNBT(tag, instance.fuelInv);
                }
            }
        }, CapabilityMinecartDestination::new);
    }

    @Override
    public void setText(int textFieldID, String text){
        destinationStations = text;
        recompileRegexes();
    }

    @Override
    public void setDestinations(String... destinations){
        Validate.noNullElements(destinations, "The destinations array contains null at position %d");

        destinationStations = StringUtils.joinWith("\n", (Object[])destinations);
        recompileRegexes();
    }

    private void recompileRegexes(){
        String[] destinations = getDestinations();
        destinationRegexes = new Pattern[destinations.length];
        invalidDestinations = "";
        for(int i = 0; i < destinations.length; i++) {
            try {
                destinationRegexes[i] = Pattern.compile(destinations[i]);
            } catch(PatternSyntaxException e) {
                if(!invalidDestinations.equals("")) {
                    invalidDestinations += ",";
                }
                invalidDestinations += "" + i;
                destinationRegexes[i] = EMPTY_PATTERN;
            }
        }
        getCurrentDestination(); //Update to a valid destination index.
    }

    @Override
    public int[] getInvalidDestinationIndeces(){
        if(invalidDestinations.equals("")) return new int[0];
        String[] strings = invalidDestinations.split(",");
        int[] ints = new int[strings.length];
        for(int i = 0; i < strings.length; i++) {
            ints[i] = Integer.parseInt(strings[i]);
        }
        return ints;
    }

    @Override
    public String getText(int textFieldID){
        return destinationStations;
    }

    public String getDestination(int index){
        return getDestinations()[index];
    }

    @Override
    public String[] getDestinations(){
        return destinationStations.equals("") ? new String[0] : destinationStations.split("\n");
    }

    @Override
    public int getTotalDestinations(){
        return getDestinations().length;
    }

    @Override
    public String getCurrentDestination(){
        String[] destinations = getDestinations();
        if(curDestinationIndex >= destinations.length || curDestinationIndex == -1) nextDestination();
        return curDestinationIndex >= 0 ? destinations[curDestinationIndex] : "";
    }

    @Override
    public int getDestinationIndex(){
        return curDestinationIndex;
    }

    public void nextDestination(){
        setCurrentDestinationIndex(curDestinationIndex + 1);
    }

    @Override
    public void setCurrentDestinationIndex(int index){
        String[] destinations = getDestinations();
        if(index >= destinations.length || index < 0) {
            curDestinationIndex = destinations.length > 0 ? 0 : -1;
        } else {
            curDestinationIndex = index;
        }
    }

    public Pattern getCurrentDestinationRegex(){
        getCurrentDestination();
        return curDestinationIndex >= 0 ? destinationRegexes[curDestinationIndex] : EMPTY_PATTERN;
    }

    public void setPath(EntityMinecart cart, AStarRailNode path){
        curPath = path;
        nbtLoadedPath = null;
        sendUpdatePacket(cart);
    }

    private static void sendUpdatePacket(EntityMinecart cart){
        NetworkHandler.sendToAll(new PacketUpdateMinecartPath(cart));
    }

    public AStarRailNode getPath(World world){
        if(nbtLoadedPath != null) {
            AStarRailNode prevNode = null;
            for(int i = nbtLoadedPath.size() - 1; i >= 0; i--) {
                AStarRailNode curNode = new AStarRailNode(RailCacheManager.getInstance(world).getRail(world, nbtLoadedPath.get(i)), null, null);
                if(prevNode != null) curNode.checkImprovementAndUpdate(prevNode);
                prevNode = curNode;
            }
            curPath = prevNode;
            nbtLoadedPath = null;
        }
        return curPath;
    }

    public List<BlockPos> getNBTPath(){
        return nbtLoadedPath;
    }

    public void setMotorized(){
        motorized = true;
    }

    public boolean isMotorized(){
        return motorized;
    }

    /**
     * Tries to use fuel and returns true if succeeded.
     * @return
     */
    public boolean useFuel(EntityMinecart cart){
        if(motorized) {
            if(fuelLeft == 0) {
                for(int i = 0; i < fuelInv.func_70302_i_(); i++) {
                    ItemStack fuel = fuelInv.func_70301_a(i);
                    if(!fuel.func_190926_b()) {
                        int fuelValue = TileEntityFurnace.func_145952_a(fuel);
                        if(fuelValue > 0) {
                            fuel.func_190918_g(1);
                            if(fuel.func_190926_b()) {
                                fuelInv.func_70299_a(i, fuel.func_77973_b().getContainerItem(fuel));
                            }
                            fuelLeft += fuelValue;
                            totalBurnTime = fuelValue;
                            break;
                        }
                    }
                }
            }
            if(fuelLeft > 0) {
                fuelLeft--;
                double randX = cart.func_174791_d().field_72450_a + (cart.field_70170_p.field_73012_v.nextDouble() - 0.5) * 0.5;
                double randY = cart.func_174791_d().field_72448_b + (cart.field_70170_p.field_73012_v.nextDouble() - 0.5) * 0.5;
                double randZ = cart.func_174791_d().field_72449_c + (cart.field_70170_p.field_73012_v.nextDouble() - 0.5) * 0.5;
                NetworkHandler.sendToAllAround(new PacketSpawnParticle(EnumParticleTypes.SMOKE_LARGE, randX, randY, randZ, 0, 0, 0), cart.field_70170_p);
                return true;
            }
        }
        return false;
    }

    public InventoryBasic getFuelInv(){
        return fuelInv;
    }

    public IItemHandler getEngineItemHandler(){
        return motorized ? fuelItemHandler : engineItemHandler;
    }

    public int getScaledFuel(int barLength){
        return totalBurnTime == 0 ? 0 : barLength * fuelLeft / totalBurnTime;
    }

    public void onCartBroken(EntityMinecart cart){
        if(motorized && !travelingBetweenDimensions) {
            motorized = false;
            cart.func_145779_a(ModItems.CART_ENGINE, 1);
            for(int i = 0; i < fuelInv.func_70302_i_(); i++) {
                ItemStack fuel = fuelInv.func_70301_a(i);
                if(fuel != null) cart.func_70099_a(fuel, 0);
            }
        }
        travelingBetweenDimensions = false;
    }

    public void setEngineActive(boolean active){
        motorActive = active;
    }

    public void onCartUpdate(MinecartUpdateEvent event){
        EntityMinecart cart = event.getMinecart();
        if(!cart.field_70170_p.field_72995_K) {
            if(isMotorized()) {
                boolean shouldRun = true;
                EnumFacing cartDir = cart.func_184172_bi();
                if(new Vec3d(cart.field_70159_w, cart.field_70181_x, cart.field_70179_y).func_72433_c() < 0.05) {
                    shouldRun = false;
                    if(hopperTimer > 0) {
                        hopperTimer--;
                    }
                    if(hopperTimer == 0) {
                        hopperTimer = extractFuelFromHopper(cart, event.getPos()) ? 8 : 40;
                    }
                } else {
                    hopperTimer = 0;
                    RailWrapper rail = RailCacheManager.getInstance(cart.field_70170_p).getRail(cart.field_70170_p, event.getPos());

                    if(rail == null) {
                        shouldRun = false;
                    } else {
                        TileEntitySignalBase signal = TileEntitySignalBase.getNeighborSignal(rail, cartDir.func_176734_d());
                        shouldRun = signal == null || signal.getLampStatus() != EnumLampStatus.RED;
                        if(!shouldRun) {
                            cart.field_70159_w = 0;
                            cart.field_70179_y = 0;
                        }
                    }
                }

                if(shouldRun && useFuel(cart)) {
                    if(!motorActive) NetworkHandler.sendToAllAround(new PacketUpdateMinecartEngineState(cart, true), new NetworkRegistry.TargetPoint(cart.field_70170_p.field_73011_w.getDimension(), cart.func_174791_d().field_72450_a, cart.func_174791_d().field_72448_b, cart.func_174791_d().field_72449_c, 64));
                    motorActive = true;

                    double acceleration = 0.03D;
                    cart.field_70159_w += cartDir.func_82601_c() * acceleration;
                    cart.field_70179_y += cartDir.func_82599_e() * acceleration;
                    cart.field_70159_w = MathHelper.func_151237_a(cart.field_70159_w, -cart.getMaxCartSpeedOnRail(), cart.getMaxCartSpeedOnRail());
                    cart.field_70179_y = MathHelper.func_151237_a(cart.field_70179_y, -cart.getMaxCartSpeedOnRail(), cart.getMaxCartSpeedOnRail());
                } else {
                    if(motorActive) NetworkHandler.sendToAllAround(new PacketUpdateMinecartEngineState(cart, true), new NetworkRegistry.TargetPoint(cart.field_70170_p.field_73011_w.getDimension(), cart.func_174791_d().field_72450_a, cart.func_174791_d().field_72448_b, cart.func_174791_d().field_72449_c, 64));
                    motorActive = false;
                }
            }
        } else {
            if(motorActive) {
                cart.field_70170_p.func_175688_a(EnumParticleTypes.SMOKE_LARGE, cart.func_174791_d().field_72450_a, cart.func_174791_d().field_72448_b, cart.func_174791_d().field_72449_c, 0, 0, 0);
            }
        }
    }

    /**
     * 
     * @param cart
     * @return true if there was a valid hopper (not necessarily if extracted an item)
     */
    private boolean extractFuelFromHopper(EntityMinecart cart, BlockPos pos){
        boolean foundHopper = false;
        for(EnumFacing dir : EnumFacing.field_82609_l) {
            BlockPos neighbor = pos;
            for(int offsetTimes = 0; offsetTimes < (dir == EnumFacing.UP ? 2 : 1); offsetTimes++) {
                neighbor = neighbor.func_177972_a(dir);
                TileEntity te = cart.field_70170_p.func_175625_s(neighbor);
                if(te instanceof TileEntityHopper) {
                    EnumFacing hopperDir = cart.field_70170_p.func_180495_p(neighbor).func_177229_b(BlockHopper.field_176430_a);
                    if(hopperDir.func_176734_d() == dir) {
                        TileEntityHopper hopper = (TileEntityHopper)te;
                        for(int i = 0; i < hopper.func_70302_i_(); i++) {
                            ItemStack stack = hopper.func_70301_a(i);
                            if(!stack.func_190926_b() && getFuelInv().func_94041_b(0, stack)) {
                                ItemStack inserted = stack.func_77946_l();
                                inserted.func_190920_e(1);
                                ItemStack left = ItemHandlerHelper.insertItemStacked(getEngineItemHandler(), inserted, false);
                                if(left.func_190926_b()) {
                                    stack.func_190918_g(1);
                                    hopper.func_70296_d();
                                    return true;
                                }
                            }
                        }
                        foundHopper = true;
                    }
                }
            }
        }
        return foundHopper;
    }

    public static class Provider implements ICapabilitySerializable<NBTBase>{
        private final CapabilityMinecartDestination cap = new CapabilityMinecartDestination();

        @Override
        public boolean hasCapability(Capability<?> capability, EnumFacing facing){
            return capability == INSTANCE;
        }

        @Override
        public <T> T getCapability(Capability<T> capability, EnumFacing facing){
            if(hasCapability(capability, facing)) {
                return (T)cap;
            } else {
                return null;
            }
        }

        @Override
        public NBTBase serializeNBT(){
            return INSTANCE.getStorage().writeNBT(INSTANCE, cap, null);
        }

        @Override
        public void deserializeNBT(NBTBase nbt){
            INSTANCE.getStorage().readNBT(INSTANCE, cap, null, nbt);
        }
    }
}
