package com.minemaarten.signals.client.render.signals;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import net.minecraft.block.BlockRailBase.EnumRailDirection;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.item.ItemDye;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;

import org.lwjgl.opengl.GL11;

import com.minemaarten.signals.client.RectRenderer;
import com.minemaarten.signals.lib.SignalBlockNode;
import com.minemaarten.signals.lib.SignalBlockSection;
import com.minemaarten.signals.lib.SignalsUtils;
import com.minemaarten.signals.lib.Vec3iUtils;
import com.minemaarten.signals.tileentity.TileEntitySignalBase;

public class BlockSectionRenderer{

    private Map<TileEntitySignalBase, SignalBlockSectionColored> signalsToColors = new HashMap<>();
    private int refreshCounter = 0;

    private SignalBlockSectionColored getBlockSection(TileEntitySignalBase te){
        SignalBlockSectionColored blockSection = signalsToColors.get(te);
        if(blockSection == null) {
            SignalBlockSectionColored newSection = new SignalBlockSectionColored(te.getSignalBlockInfo());
            Optional<SignalBlockSectionColored> matchingSection = signalsToColors.values().stream().filter(x -> x.equals(newSection)).findFirst();
            if(matchingSection.isPresent()) {
                blockSection = matchingSection.get();
            } else {
                blockSection = newSection;

                Set<Integer> invalidColors = getAdjacentBlockSections(blockSection).map(x -> x.colorIndex).collect(Collectors.toSet());
                int availableColors = 16 - invalidColors.size();
                if(availableColors > 0) { //If there are colors left (it would be very exceptional if there weren't.
                    int usedIndex = blockSection.getRootNode().hashCode() % availableColors; //Use a deterministic way to generate a color index.
                    for(int i = 0; i < 16; i++) {
                        if(!invalidColors.contains(i)) {
                            if(usedIndex-- <= 0) {
                                blockSection.colorIndex = i;
                                break;
                            }
                        }
                    }
                }

                blockSection.compileRender();
            }

            signalsToColors.put(te, blockSection);
        }
        return blockSection;
    }

    private Stream<SignalBlockSectionColored> getAdjacentBlockSections(SignalBlockSectionColored blockSection){
        return signalsToColors.values().stream().filter(otherSection -> otherSection.isAdjacent(blockSection));
    }

    public void render(BufferBuilder b, List<TileEntity> tes){
        if(refreshCounter++ >= 100) {
            refreshCounter = 0;
            signalsToColors.clear();

            for(TileEntity te : tes) {
                if(te instanceof TileEntitySignalBase) {
                    TileEntitySignalBase teSignal = (TileEntitySignalBase)te;
                    getBlockSection(teSignal);
                }
            }
        }

        b.func_181668_a(GL11.GL_TRIANGLES, DefaultVertexFormats.field_181706_f);
        for(TileEntity te : tes) {
            if(te instanceof TileEntitySignalBase) {
                TileEntitySignalBase teSignal = (TileEntitySignalBase)te;
                renderSignalDirection(b, teSignal);
            }
        }
        Tessellator.func_178181_a().func_78381_a();

        b.func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_181706_f);
        for(SignalBlockSectionColored blockSection : signalsToColors.values()) {
            blockSection.rectRenderer.render(b);
        }
        Tessellator.func_178181_a().func_78381_a();
    }

    private int getRailHeightOffset(EnumRailDirection railDir, EnumFacing dir){
        switch(railDir){
            case ASCENDING_EAST:
                return dir == EnumFacing.EAST ? 1 : (dir == EnumFacing.WEST ? -1 : 0);
            case ASCENDING_NORTH:
                return dir == EnumFacing.NORTH ? 1 : (dir == EnumFacing.SOUTH ? -1 : 0);
            case ASCENDING_SOUTH:
                return dir == EnumFacing.SOUTH ? 1 : (dir == EnumFacing.NORTH ? -1 : 0);
            case ASCENDING_WEST:
                return dir == EnumFacing.WEST ? 1 : (dir == EnumFacing.EAST ? -1 : 0);
            default:
                return 0;
        }
    }

    private void renderSignalDirection(BufferBuilder buffer, TileEntitySignalBase signal){
        EnumFacing signalFacing = signal.getFacing().func_176734_d();
        SignalBlockNode rootNode = signal.getSignalBlockInfo();
        int heightOffset = getRailHeightOffset(rootNode.railDir, signalFacing);

        EnumFacing rotatedFacing = signalFacing.func_176746_e();
        EnumFacing rotatedFacing2 = signalFacing.func_176735_f();

        int colorIndex = getBlockSection(signal).colorIndex;
        int color = ItemDye.field_150922_c[colorIndex];
        float r = (color >> 16) / 256F;
        float g = (color >> 8 & 255) / 256F;
        float b = (color & 255) / 256F;

        BlockPos pos = signal.func_174877_v().func_177972_a(rotatedFacing);
        Vec3d posVec = new Vec3d(pos.func_177958_n() + 0.5, pos.func_177956_o() + (heightOffset != 0 ? 0.6 : 0.1), pos.func_177952_p() + 0.5);

        double yOffset = heightOffset * 1;

        double arrowSize = 0.2;
        double spacing = 0.2;
        for(int i = -2; i < 0; i++) {
            Vec3d shiftedPosVec = posVec.func_72441_c(signalFacing.func_82601_c() * spacing * i, spacing * i * yOffset, signalFacing.func_82599_e() * spacing * i);
            Vec3d vecBack = shiftedPosVec.func_72441_c(signalFacing.func_82601_c() * arrowSize, arrowSize * yOffset, signalFacing.func_82599_e() * arrowSize);
            Vec3d c1 = vecBack.func_72441_c(rotatedFacing.func_82601_c() * arrowSize, 0, rotatedFacing.func_82599_e() * arrowSize);
            Vec3d c2 = vecBack.func_72441_c(rotatedFacing2.func_82601_c() * arrowSize, 0, rotatedFacing2.func_82599_e() * arrowSize);

            buffer.func_181662_b(shiftedPosVec.field_72450_a, shiftedPosVec.field_72448_b, shiftedPosVec.field_72449_c).func_181666_a(r, g, b, 1).func_181675_d();
            buffer.func_181662_b(c1.field_72450_a, c1.field_72448_b, c1.field_72449_c).func_181666_a(r, g, b, 1).func_181675_d();
            //buffer.pos(shiftedPosVec.x, shiftedPosVec.y, shiftedPosVec.z).color(r, g, b, 1).endVertex();
            buffer.func_181662_b(c2.field_72450_a, c2.field_72448_b, c2.field_72449_c).func_181666_a(r, g, b, 1).func_181675_d();
        }
    }

    private void renderSignalBlocks(RectRenderer rectRenderer, SignalBlockNode node, boolean goingDown){
        for(SignalBlockNode neighbor : node.nextNeighbors) {
            boolean isLowerNeighbor = neighbor.railPos.func_177956_o() < node.railPos.func_177956_o();
            boolean isHigherNeighbor = neighbor.railPos.func_177956_o() > node.railPos.func_177956_o();

            Vec3d interpolated = Vec3iUtils.interpolate(node.railPos, neighbor.railPos);
            EnumFacing dir = SignalsUtils.getRelativeHorizonal(node.railPos, neighbor.railPos);

            if(dir != null) { //When adjacent
                boolean neighborsNeighborIsHigher = getRailHeightOffset(neighbor.railDir, dir) == 1;

                rectRenderer.pos(node.railPos.func_177958_n() + 0.5, node.railPos.func_177956_o() + (isHigherNeighbor || goingDown ? 0.6 : 0.1), node.railPos.func_177952_p() + 0.5);
                rectRenderer.pos(interpolated.field_72450_a + 0.5, node.railPos.func_177956_o() + (isHigherNeighbor ? 1.1 : 0.1), interpolated.field_72449_c + 0.5);
                rectRenderer.pos(interpolated.field_72450_a + 0.5, node.railPos.func_177956_o() + (isHigherNeighbor ? 1.1 : 0.1), interpolated.field_72449_c + 0.5);
                rectRenderer.pos(neighbor.railPos.func_177958_n() + 0.5, neighbor.railPos.func_177956_o() + (isLowerNeighbor || neighborsNeighborIsHigher ? 0.6 : 0.1), neighbor.railPos.func_177952_p() + 0.5);
            } else { //When not adjacent (Rail Link, for example)
                isLowerNeighbor = false;
            }

            renderSignalBlocks(rectRenderer, neighbor, isLowerNeighbor);
        }
    }

    private class SignalBlockSectionColored extends SignalBlockSection{

        public int colorIndex;
        public RectRenderer rectRenderer;

        public SignalBlockSectionColored(SignalBlockNode rootNode){
            super(rootNode);
        }

        public void compileRender(){
            rectRenderer = new RectRenderer();

            int color = ItemDye.field_150922_c[colorIndex];
            float r = (color >> 16) / 256F;
            float g = (color >> 8 & 255) / 256F;
            float b = (color & 255) / 256F;
            rectRenderer.setColor(r, g, b);
            SignalBlockNode rootNode = getRootNode();
            boolean goingDown = false;

            //Determine going down for the first node
            if(!rootNode.nextNeighbors.isEmpty()) {
                SignalBlockNode neighbor = rootNode.nextNeighbors.get(0); //The first node can only have one neighbor.
                EnumFacing relDir = SignalsUtils.getRelativeHorizonal(rootNode.railPos, neighbor.railPos);
                goingDown = getRailHeightOffset(rootNode.railDir, relDir) == -1;
            }

            renderSignalBlocks(rectRenderer, rootNode, goingDown);
        }
    }
}
