package com.minemaarten.signals.rail;

import java.util.Arrays;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.Gui;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.texture.DynamicTexture;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.common.DimensionManager;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import org.lwjgl.opengl.GL11;

import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.inventory.ContainerNetworkController;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketUpdateNetworkController;
import com.minemaarten.signals.tileentity.TileEntitySignalBase;
import com.minemaarten.signals.tileentity.TileEntityStationMarker;

public class NetworkController{
    private static final Map<Integer, NetworkController> SERVER_INSTANCES = new HashMap<>();
    private static final Map<Integer, NetworkController> CLIENT_INSTANCES = new HashMap<>();
    private static final int RAIL_COLOR = 0xFF666666;
    private static final int PATH_COLOR = 0xFFAAAAAA;
    private static final int TEXT_COLOR = 0xFFFFFF; //No alpha
    private static final int STATION_COLOR = 0xFFDDDD00;
    private static final int NOTHING_COLOR = 0xFF222222;

    public static NetworkController getInstance(World world){
        return getInstance(world.field_73011_w.getDimension(), world.field_72995_K);
    }

    public static NetworkController getInstance(int dimension, boolean client){
        Map<Integer, NetworkController> cache = client ? CLIENT_INSTANCES : SERVER_INSTANCES;
        NetworkController controller = cache.get(dimension);
        if(controller == null) {
            controller = new NetworkController(dimension);
            if(client) controller.setColors(new int[]{0}, 1, 1, 0, 0);
            cache.put(dimension, controller);
            if(!client) controller.rebuildAll();
        }
        return controller;
    }

    @SideOnly(Side.CLIENT)
    private DynamicTexture dynamicTexture;
    private int width, height;
    private int[] colors;
    private int startX, startZ;
    private final int dimensionId;
    private ResourceLocation textureLoc;

    public NetworkController(int dimensionId){
        this.dimensionId = dimensionId;
    }

    @SideOnly(Side.CLIENT)
    public void setColors(int[] colors, int width, int height, int startX, int startZ){
        this.startX = startX;
        this.startZ = startZ;
        if(dynamicTexture == null || this.width != width || this.height != height) {
            this.width = width;
            this.height = height;
            if(dynamicTexture != null) dynamicTexture.func_147631_c();
            dynamicTexture = new DynamicTexture(width, height);
            textureLoc = Minecraft.func_71410_x().func_110434_K().func_110578_a("networkController/" + dimensionId, dynamicTexture);
        }
        System.arraycopy(colors, 0, dynamicTexture.func_110565_c(), 0, colors.length);
        dynamicTexture.func_110564_a();
    }

    public int getWidth(){
        return width;
    }

    public int getHeight(){
        return height;
    }

    public int getStartX(){
        return startX;
    }

    public int getStartZ(){
        return startZ;
    }

    @SideOnly(Side.CLIENT)
    public void render(World world){
        if(textureLoc == null) setColors(new int[1], 1, 1, 0, 0);
        Minecraft mc = Minecraft.func_71410_x();
        mc.func_110434_K().func_110577_a(textureLoc);
        GL11.glEnable(GL11.GL_TEXTURE_2D);
        GL11.glColor4f(1.0F, 1.0F, 1.0F, 1.0F);
        Gui.func_146110_a(0, 0, 0, 0, width, height, width, height);
        Tessellator t = Tessellator.func_178181_a();
        BufferBuilder buffer = t.func_178180_c();
        for(EntityMinecart cart : world.func_72872_a(EntityMinecart.class, new AxisAlignedBB(startX, 0, startZ, startX + width, 255, startZ + height))) {
            List<BlockPos> path = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null).getNBTPath();
            if(path != null) {
                for(BlockPos pathPos : path) {
                    int pathX = pathPos.func_177958_n() - startX;
                    int pathZ = pathPos.func_177952_p() - startZ;
                    Gui.func_73734_a(pathX, pathZ, pathX + 1, pathZ + 1, PATH_COLOR);
                }
            }
        }
        for(EntityMinecart cart : world.func_72872_a(EntityMinecart.class, new AxisAlignedBB(startX, 0, startZ, startX + width, 255, startZ + height))) {
            double x = cart.field_70165_t - startX - 0.5;
            double y = cart.field_70161_v - startZ - 0.5;
            GlStateManager.func_179124_c(0, 0, 1);
            GlStateManager.func_179090_x();
            buffer.func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_181705_e);
            buffer.func_181662_b(x, y, 0).func_181675_d();
            buffer.func_181662_b(x, y + 1, 0).func_181675_d();
            buffer.func_181662_b(x + 1, y + 1, 0).func_181675_d();
            buffer.func_181662_b(x + 1, y, 0).func_181675_d();
            t.func_78381_a();
            GlStateManager.func_179098_w();

            String dest = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null).getCurrentDestination();
            if(dest != null) {
                GlStateManager.func_179094_E();
                double scale = 1 / 4D;
                GlStateManager.func_179137_b(x + 1.5, y, 0);
                GlStateManager.func_179139_a(scale, scale, scale);
                mc.field_71466_p.func_78276_b(dest, 0, 0, TEXT_COLOR);
                GlStateManager.func_179121_F();
            }
        }
        for(TileEntity te : world.field_147482_g) {
            if(te instanceof TileEntityStationMarker) {
                double x = te.func_174877_v().func_177958_n() - startX - 0.5;
                double y = te.func_174877_v().func_177952_p() - startZ - 0.5;

                GlStateManager.func_179094_E();
                double scale = 1 / 4D;
                GlStateManager.func_179137_b(x + 2, y, 0);
                GlStateManager.func_179139_a(scale, scale, scale);
                mc.field_71466_p.func_78276_b(((TileEntityStationMarker)te).getStationName(), 0, 0, 0xFFFFFF00);
                GlStateManager.func_179121_F();
            }
        }
    }

    private void setColor(int x, int z, int color){
        setColor(x, z, color, true);
    }

    private void setColor(int x, int z, int color, boolean sendPacket){
        int index = (x - startX) + (z - startZ) * width;
        if(index < 0 || index >= colors.length) { //If we are out of bounds that means we don't properly know the right dims.
            rebuildAll();
        } else if(colors[index] != color) {
            colors[index] = color;
            if(sendPacket) sendUpdatePacket();
        }
    }

    private void sendUpdatePacket(){
        PacketUpdateNetworkController packet = new PacketUpdateNetworkController(dimensionId, colors, width, height, startX, startZ);
        for(EntityPlayer player : DimensionManager.getWorld(dimensionId).field_73010_i) {
            if(shouldPlayerGetUpdates(player)) {
                NetworkHandler.sendTo(packet, (EntityPlayerMP)player);
            }
        }

    }

    private static boolean shouldPlayerGetUpdates(EntityPlayer player){
        return player.field_71070_bA instanceof ContainerNetworkController;
    }

    public void updateColor(RailWrapper rail, BlockPos pos){
        updateColor(rail != null ? RAIL_COLOR : NOTHING_COLOR, pos);
    }

    public void updateColor(TileEntitySignalBase signal, BlockPos pos){
        updateColor(signal != null ? signal.getLampStatus().color : NOTHING_COLOR, pos);
    }

    public void updateColor(TileEntityStationMarker station, BlockPos pos){
        updateColor(station != null ? STATION_COLOR : NOTHING_COLOR, pos);
    }

    public void updateColor(int color, BlockPos... positions){
        boolean rebuildAll = false;
        for(BlockPos pos : positions) {
            if(color == NOTHING_COLOR) {
                if(pos.func_177958_n() == startX || pos.func_177958_n() == startX + width || pos.func_177952_p() == startZ || pos.func_177952_p() == startZ + height) {
                    rebuildAll = true; //When a border rail is removed it might decrease the total map size.
                }
            } else {
                if(pos.func_177958_n() < startX || pos.func_177958_n() >= startX + width || pos.func_177952_p() < startZ || pos.func_177952_p() >= startZ + height) {
                    rebuildAll = true; //When the rail is outside the current map scope we need to increase the map size.
                }
            }
            if(!rebuildAll) setColor(pos.func_177958_n(), pos.func_177952_p(), color);
        }
        if(rebuildAll) rebuildAll();
    }

    public void rebuildAll(){
        Iterable<RailWrapper> allRails = RailCacheManager.getInstance(dimensionId).getAllRails();
        Map<BlockPos, Integer> posToColor = new HashMap<>();
        for(RailWrapper wrapper : allRails) {
            posToColor.put(wrapper, RAIL_COLOR);
        }
        try {
            for(TileEntity te : DimensionManager.getWorld(dimensionId).field_175730_i) {
                if(te instanceof TileEntitySignalBase) {
                    posToColor.put(te.func_174877_v(), ((TileEntitySignalBase)te).getLampStatus().color);
                } else if(te instanceof TileEntityStationMarker) {
                    posToColor.put(te.func_174877_v(), STATION_COLOR);
                }
            }
        } catch(ConcurrentModificationException e) {

        }

        //allRails = Lists.newArrayList(allRails);
        startX = Integer.MAX_VALUE;
        startZ = Integer.MAX_VALUE;
        int endX = Integer.MIN_VALUE;
        int endZ = Integer.MIN_VALUE;
        for(BlockPos pos : posToColor.keySet()) {
            startX = Math.min(startX, pos.func_177958_n());
            startZ = Math.min(startZ, pos.func_177952_p());
            endX = Math.max(endX, pos.func_177958_n());
            endZ = Math.max(endZ, pos.func_177952_p());
        }

        if(posToColor.isEmpty()) {
            width = 1;
            height = 1;
            colors = new int[1];
            Arrays.fill(colors, NOTHING_COLOR);
        } else {
            width = endX - startX + 1;
            height = endZ - startZ + 1;
            colors = new int[width * height];
            Arrays.fill(colors, NOTHING_COLOR);
            for(Map.Entry<BlockPos, Integer> entry : posToColor.entrySet()) {
                setColor(entry.getKey().func_177958_n(), entry.getKey().func_177952_p(), entry.getValue(), false);
            }
        }

        sendUpdatePacket();
    }
}
