package com.minemaarten.signals.tileentity;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.util.math.BlockPos;

import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.lib.Log;
import com.minemaarten.signals.rail.DestinationPathFinder.AStarRailNode;
import com.minemaarten.signals.rail.RailWrapper;

import com.minemaarten.signals.api.access.ISignal.EnumForceMode;
import com.minemaarten.signals.api.access.ISignal.EnumLampStatus;

public class TileEntityPathSignal extends TileEntitySignalBase{

    private int pathingTimer;

    @Override
    public void func_73660_a(){
        super.func_73660_a();
        if(!field_145850_b.field_72995_K && pathingTimer-- <= 0 && getForceMode() == EnumForceMode.NONE) {
            route();
        }
    }

    public void route(){
        pathingTimer = 20;
        RailWrapper neighborRail = getConnectedRail();
        if(neighborRail != null) {
            List<EntityMinecart> routingMinecarts = getNeighborMinecarts();
            if(!routingMinecarts.isEmpty()) {
                Set<RailWrapper> rails = getRailsToNextBlockSection(neighborRail, getFacing());
                List<EntityMinecart> cartsOnNextBlock = getMinecarts(field_145850_b, rails);

                //Fall back onto default Block Signal behaviour if there are any carts to be routed without a path. 
                //This also will fill the routed carts with a path for the next stage.
                for(EntityMinecart routingCart : routingMinecarts) {
                    if(routeCart(routingCart, getFacing(), false) == null) {
                        setLampStatus(cartsOnNextBlock.isEmpty() ? EnumLampStatus.GREEN : EnumLampStatus.RED);
                        setMessage("signals.signal_message.cart_without_destination");
                        Log.debug("[Path Signal] Cart routed without destination. Block signal behaviour.");
                        return;
                    }
                }

                //Don't allow the cart to proceed if there are carts on the block without a path.
                for(EntityMinecart cartOnNextBlock : cartsOnNextBlock) {
                    if(getStoredPath(cartOnNextBlock) == null) {
                        setLampStatus(EnumLampStatus.RED);
                        BlockPos pos = cartOnNextBlock.func_180425_c();
                        setMessage("signals.signal_message.cart_on_track_without_destination", pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p());
                        Log.debug("[Path Signal] Cart on rails without destination. Red signal. Cart: " + cartOnNextBlock.func_180425_c());
                        return;
                    }
                }

                //The positions of the track that should be kept free of carts for the carts to be allowed to continue.
                Set<BlockPos> claimingPositions = new HashSet<BlockPos>();
                for(EntityMinecart routingCart : routingMinecarts) {
                    claimingPositions.addAll(getToBeTraversedCoordinates(routingCart));
                }

                for(EntityMinecart cartOnNextBlock : cartsOnNextBlock) {
                    List<BlockPos> list = getToBeTraversedCoordinates(cartOnNextBlock);
                    for(BlockPos pos : list) {
                        if(claimingPositions.contains(pos)) {
                            setLampStatus(EnumLampStatus.RED);
                            BlockPos p = cartOnNextBlock.func_180425_c();
                            setMessage("signals.signal_message.cart_intersecting_path", p.func_177958_n(), p.func_177956_o(), p.func_177952_p());
                            Log.debug("[Path Signal] Cart on rails intersecting the path of the routed cart. Red signal. Cart: " + cartOnNextBlock.func_180425_c());
                            return;
                        }
                    }
                }
                setLampStatus(EnumLampStatus.GREEN);

                AStarRailNode path = getStoredPath(routingMinecarts.get(0));
                if(path != null) updateSwitches(path, routingMinecarts.get(0), true);
            } else {
                setMessage("");
                // setMessage("signals.signal_message.standby");
                setLampStatus(EnumLampStatus.YELLOW);
            }
        } else {
            setLampStatus(EnumLampStatus.YELLOW_BLINKING);
        }
    }

    @Override
    protected void onCartEnteringBlock(EntityMinecart cart){
        route();
    }

    private static AStarRailNode getStoredPath(EntityMinecart cart){
        return cart.getCapability(CapabilityMinecartDestination.INSTANCE, null).getPath(cart.field_70170_p);
    }

    private static List<BlockPos> getToBeTraversedCoordinates(EntityMinecart cart){
        AStarRailNode path = getStoredPath(cart);
        List<BlockPos> coords = new ArrayList<>();
        BlockPos cartPos = cart.func_180425_c();
        boolean returnOnNext = false;

        if(path != null) path = path.getNextNode();

        Stack<AStarRailNode> reversedRoute = new Stack<>();
        for(; path != null; path = path.getNextNode()) {
            if(path.getSignal(null) != null) break;
            reversedRoute.push(path);
        }

        while(!reversedRoute.isEmpty()) {
            AStarRailNode node = reversedRoute.pop();

            coords.add(node.getRail());

            if(returnOnNext) return coords;
            if(node.getRail().equals(cartPos)) returnOnNext = true;
        }
        return coords;
    }

}
