package com.minemaarten.signals.tileentity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import net.minecraft.block.Block;
import net.minecraft.block.BlockRailBase.EnumRailDirection;
import net.minecraft.block.state.IBlockState;
import net.minecraft.client.resources.I18n;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ITickable;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;

import com.minemaarten.signals.api.access.ISignal;
import com.minemaarten.signals.block.BlockSignalBase;
import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.lib.Log;
import com.minemaarten.signals.lib.SignalBlockNode;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketUpdateMessage;
import com.minemaarten.signals.rail.DestinationPathFinder;
import com.minemaarten.signals.rail.DestinationPathFinder.AStarRailNode;
import com.minemaarten.signals.rail.NetworkController;
import com.minemaarten.signals.rail.RailCacheManager;
import com.minemaarten.signals.rail.RailManager;
import com.minemaarten.signals.rail.RailWrapper;
import com.minemaarten.signals.rail.SignalsOnRouteIterable.SignalOnRoute;

import com.minemaarten.signals.api.access.ISignal.EnumForceMode;
import com.minemaarten.signals.api.access.ISignal.EnumLampStatus;

public abstract class TileEntitySignalBase extends TileEntityBase implements ITickable, ISignal{

    private boolean firstTick = true;
    private List<EntityMinecart> routedMinecarts = new ArrayList<>();
    private Set<TileEntitySignalBase> nextSignals = new HashSet<>();
    private SignalBlockNode rootSignalNode = new SignalBlockNode(new BlockPos(0, 0, 0), EnumRailDirection.NORTH_SOUTH);
    private String text = "";
    private String arguments = "";
    private EnumForceMode forceMode = EnumForceMode.NONE;
    private EntityMinecart claimingCart; //The cart that has called dibs on the rail block in front of this signal.
    private UUID claimingCartUUID; //The claiming cart ID loaded from NBT. Will only have a value when just loaded.

    public BlockPos getNeighborPos(){
        return func_174877_v().func_177972_a(getFacing().func_176735_f());
    }

    public RailWrapper getConnectedRail(){
        BlockPos neighborPos = getNeighborPos();
        RailWrapper rail = RailCacheManager.getInstance(func_145831_w()).getRail(func_145831_w(), neighborPos);
        return rail != null && rail.isStraightTrack() && rail.getNeighbors().size() <= 2 ? rail : null;
    }

    public List<RailWrapper> getConnectedRails(){
        RailWrapper neighbor = getConnectedRail();
        return neighbor != null ? getConnectedRails(neighbor, getFacing().func_176734_d()) : new ArrayList<>();
    }

    public static List<RailWrapper> getConnectedRails(RailWrapper neighbor, EnumFacing traverseDirection){
        List<RailWrapper> neighbors = new ArrayList<>();
        Block railType = neighbor.state.func_177230_c();
        for(int i = 0; i < 5 && neighbor != null; i++) {
            neighbors.add(neighbor);
            Map<RailWrapper, EnumFacing> railNeighbors = neighbor.getNeighbors();
            neighbor = null;
            for(Map.Entry<RailWrapper, EnumFacing> entry : railNeighbors.entrySet()) {
                if(entry.getValue() == traverseDirection && railType == entry.getKey().state.func_177230_c()) {
                    neighbor = entry.getKey();
                    break;
                }
            }
        }
        return neighbors;
    }

    private IBlockState getBlockState(){
        return func_145831_w() != null ? func_145831_w().func_180495_p(func_174877_v()) : null;
    }

    public EnumFacing getFacing(){
        IBlockState state = getBlockState();
        return state != null && state.func_177230_c() instanceof BlockSignalBase ? state.func_177229_b(BlockSignalBase.FACING) : EnumFacing.NORTH;
    }

    public boolean isValidRoute(AStarRailNode route, EntityMinecart cart){
        return true;
    }

    protected void setLampStatus(EnumLampStatus lampStatus){
        setLampStatus(lampStatus, this::getNeighborMinecarts, cart -> routeCart(cart, getFacing(), true));
    }

    protected void setLampStatus(EnumLampStatus lampStatus, Supplier<List<EntityMinecart>> neighborMinecartGetter, Function<EntityMinecart, AStarRailNode> pathfinder){
        if(forceMode == EnumForceMode.FORCED_GREEN_ONCE) {
            lampStatus = EnumLampStatus.GREEN;
        } else if(forceMode == EnumForceMode.FORCED_RED) {
            lampStatus = EnumLampStatus.RED;
        }

        List<EntityMinecart> neighborMinecarts = null;
        if(lampStatus == EnumLampStatus.GREEN && getClaimingCart() != null) {
            neighborMinecarts = neighborMinecartGetter.get();
            lampStatus = neighborMinecarts.contains(getClaimingCart()) ? EnumLampStatus.GREEN : EnumLampStatus.YELLOW;
        }

        IBlockState state = getBlockState();
        if(state.func_177227_a().contains(BlockSignalBase.LAMP_STATUS) && state.func_177229_b(BlockSignalBase.LAMP_STATUS) != lampStatus) {
            func_145831_w().func_175656_a(func_174877_v(), state.func_177226_a(BlockSignalBase.LAMP_STATUS, lampStatus));
            NetworkController.getInstance(func_145831_w()).updateColor(this, func_174877_v());
            if(lampStatus == EnumLampStatus.GREEN) {
                //Push carts when they're standing still.
                if(neighborMinecarts == null) neighborMinecarts = neighborMinecartGetter.get();
                for(EntityMinecart cart : neighborMinecarts) {
                    if(new Vec3d(cart.field_70159_w, cart.field_70181_x, cart.field_70179_y).func_72433_c() < 0.01 || EnumFacing.func_176737_a((float)cart.field_70159_w, 0, (float)cart.field_70179_y) == getFacing()) {
                        cart.field_70159_w += getFacing().func_82601_c() * 0.1;
                        cart.field_70179_y += getFacing().func_82599_e() * 0.1;
                        long start = System.nanoTime();

                        AStarRailNode path = pathfinder.apply(cart);
                        if(path != null) updateSwitches(path, cart, true);
                        Log.debug((System.nanoTime() - start) / 1000 + "ns");
                    }
                }
            }
        }
    }

    @Override
    public EnumLampStatus getLampStatus(){
        if(func_145831_w() != null) {
            IBlockState state = func_145831_w().func_180495_p(func_174877_v());
            if(state.func_177227_a().contains(BlockSignalBase.LAMP_STATUS)) {
                return state.func_177229_b(BlockSignalBase.LAMP_STATUS);
            }
        }
        return EnumLampStatus.YELLOW_BLINKING;
    }

    protected List<EntityMinecart> getNeighborMinecarts(){
        return getMinecarts(field_145850_b, getConnectedRails());
    }

    protected AStarRailNode routeCart(EntityMinecart cart, EnumFacing cartDir, boolean submitMessages){
        CapabilityMinecartDestination capability = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null);
        String destination = capability.getCurrentDestination();
        Pattern destinationRegex = capability.getCurrentDestinationRegex();
        List<PacketUpdateMessage> messages = new ArrayList<>();
        AStarRailNode path = null;
        if(!destination.isEmpty()) {
            messages.add(new PacketUpdateMessage(this, cart, "signals.message.routing_cart", destination));
            EnumFacing facing = getFacing();
            if(facing == cartDir) {
                path = DestinationPathFinder.pathfindToDestination(getConnectedRail(), cart, destinationRegex, facing);
                if(path == null) { //If there's no path
                    messages.add(new PacketUpdateMessage(this, cart, "signals.message.no_path_found"));
                } else {
                    messages.add(new PacketUpdateMessage(this, cart, "signals.message.path_found"));
                }
            }
        } else {
            messages.add(new PacketUpdateMessage(this, cart, "signals.message.no_destination"));
        }

        if(submitMessages) {
            for(PacketUpdateMessage message : messages) {
                NetworkHandler.sendToAllAround(message, func_145831_w());
            }
        }
        capability.setPath(cart, path);

        if(destination.isEmpty()) { //When this cart is not being routed, rely on its linked carts, if any.
            return RailManager.getInstance().getPath(cart);
        } else {
            return path; //When this cart is supposed to be routed, do not rely on its linked carts.
        }
    }

    protected void updateSwitches(AStarRailNode pathNode, EntityMinecart cart, boolean submitMessages){
        List<PacketUpdateMessage> messages = new ArrayList<>();
        EnumFacing lastHeading = pathNode.getPathDir();
        while(pathNode != null) {
            Map<RailWrapper, EnumFacing> neighbors = pathNode.getRail().getNeighborsForEntryDir(lastHeading);
            EnumFacing heading = pathNode.getNextNode() != null ? neighbors.get(pathNode.getNextNode().getRail()) : null;
            if(neighbors.size() > 2 && heading != null && lastHeading != null) { //If on an intersection
                EnumRailDirection railDir = RailWrapper.getRailDir(EnumSet.of(heading, lastHeading.func_176734_d()));

                String[] args = {Integer.toString(pathNode.getRail().func_177958_n()), Integer.toString(pathNode.getRail().func_177956_o()), Integer.toString(pathNode.getRail().func_177952_p()), "signals.dir." + lastHeading.toString().toLowerCase(), "signals.dir." + heading.toString().toLowerCase()};

                if(pathNode.getRail().setRailDir(railDir)) {
                    messages.add(new PacketUpdateMessage(this, cart, "signals.message.changing_junction", args));
                } else {
                    messages.add(new PacketUpdateMessage(this, cart, "signals.message.changing_junction", args));
                }
            }
            lastHeading = heading;
            pathNode = pathNode.getNextNode();
            if(pathNode != null && heading != null && getNeighborSignal(pathNode.getRail(), heading.func_176734_d()) != null) {
                break;
            }
        }
        if(submitMessages) {
            for(PacketUpdateMessage message : messages) {
                NetworkHandler.sendToAllAround(message, func_145831_w());
            }
        }
    }

    protected static Set<RailWrapper> getRailsToNextBlockSection(RailWrapper curRail, EnumFacing direction){
        Set<RailWrapper> rails = new HashSet<>();
        Queue<Map.Entry<RailWrapper, EnumFacing>> traversingRails = new LinkedList<>();

        for(Map.Entry<RailWrapper, EnumFacing> entry : curRail.getNeighbors().entrySet()) {
            if(entry.getValue() != direction.func_176734_d()) {
                traversingRails.add(entry);
            }
        }
        rails.add(curRail); //Make sure to consider this block as traversed already, prevents traversing the tracks in reverse direction

        while(!traversingRails.isEmpty()) {
            Map.Entry<RailWrapper, EnumFacing> neighbor = traversingRails.poll();

            TileEntitySignalBase neighborSignal = getNeighborSignal(neighbor.getKey(), neighbor.getValue()); //Find a signal opposing this signal (for like merging splits)
            if(neighborSignal == null) {
                rails.add(neighbor.getKey());
                if(neighbor.getKey().getSignals().isEmpty()) {
                    //    NetworkHandler.sendToAllAround(new PacketSpawnParticle(EnumParticleTypes.REDSTONE, neighbor.getKey().getX() + 0.5, neighbor.getKey().getY() + 0.5, neighbor.getKey().getZ() + 0.5, 0, 0, 0), curRail.world);
                    for(Map.Entry<RailWrapper, EnumFacing> entry : neighbor.getKey().getNeighbors().entrySet()) {
                        BlockPos nextNeighbor = entry.getKey();
                        if(!rails.contains(nextNeighbor)) {
                            traversingRails.add(entry);
                        }
                    }
                }
            } else {
                if(curRail.world.func_82737_E() % 20 == 0) {
                    //        curRail.world.spawnEntityInWorld(new EntityFireworkRocket(curRail.world, neighborSignal.pos.getX() + 0.5, neighborSignal.pos.getY() + 0.5, neighborSignal.pos.getZ() + 0.5, new ItemStack(Items.fireworks)));
                }
                if(neighborSignal.getLampStatus() != EnumLampStatus.RED) {
                    // rails.add(neighbor.getKey());
                    // neighborSignal.setWorldObj(curRail.world); //For some reason this is necessary.
                    rails.addAll(getConnectedRails(neighbor.getKey(), neighborSignal.getFacing().func_176734_d()));
                }
            }
        }

        rails.remove(curRail); //Remove it in the end as the rail next to a signal isn't considered part of the next block.
        return rails;
    }

    @SuppressWarnings("unchecked")
    protected static <T extends TileEntitySignalBase> T getNeighborSignal(World world, BlockPos pos, Class<T> teClass, EnumFacing blacklistedSignalDir){
        for(EnumFacing dir : EnumFacing.field_176754_o) {
            BlockPos neighbor = pos.func_177972_a(dir);
            TileEntity te = world.func_175625_s(neighbor);
            if(te != null && teClass.isAssignableFrom(te.getClass())) {
                T signal = (T)te;
                if(signal.getFacing().func_176746_e() == dir && signal.getFacing() != blacklistedSignalDir) return signal;
            }
        }
        return null;
    }

    public static TileEntitySignalBase getNeighborSignal(RailWrapper rail, EnumFacing blacklistedSignalDir){
        for(Map.Entry<EnumFacing, TileEntitySignalBase> entry : rail.getSignals().entrySet()) {
            if(entry.getValue().getFacing().func_176746_e() == entry.getKey() && entry.getValue().getFacing() != blacklistedSignalDir) return entry.getValue();
        }
        return null;
    }

    public static List<EntityMinecart> getMinecarts(World worldObj, final Collection<RailWrapper> railsOnBlock){
        if(railsOnBlock.isEmpty()) return Collections.emptyList();

        Set<World> worlds = new HashSet<>();
        for(RailWrapper pos : railsOnBlock) {
            worlds.add(pos.world);
        }

        List<EntityMinecart> carts = new ArrayList<>();
        for(World world : worlds) {
            BlockPos.MutableBlockPos min = new BlockPos.MutableBlockPos(Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);
            BlockPos.MutableBlockPos max = new BlockPos.MutableBlockPos(Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE);
            for(RailWrapper pos : railsOnBlock) {
                min.func_181079_c(Math.min(min.func_177958_n(), pos.func_177958_n()), Math.min(min.func_177956_o(), pos.func_177956_o()), Math.min(min.func_177952_p(), pos.func_177952_p()));
                max.func_181079_c(Math.max(max.func_177958_n(), pos.func_177958_n()), Math.max(max.func_177956_o(), pos.func_177956_o()), Math.max(max.func_177952_p(), pos.func_177952_p()));
            }

            carts.addAll(world.func_175647_a(EntityMinecart.class, new AxisAlignedBB(min, max.func_177982_a(1, 2, 1)), cart -> {
                BlockPos cartPos = cart.func_180425_c();
                return railsOnBlock.contains(cartPos) || railsOnBlock.contains(cartPos.func_177977_b());
            }));
        }
        return carts;
    }

    protected static boolean areLinkedCartsPastTheSignal(List<EntityMinecart> routingCarts, List<EntityMinecart> cartsOnNextBlock){
        for(EntityMinecart cartOnNextBlock : cartsOnNextBlock) {
            if(isCartLinkedToAny(routingCarts, cartOnNextBlock)) {
                return true;
            }
        }
        return false;
    }

    protected static boolean isCartLinkedToAny(List<EntityMinecart> routingCarts, EntityMinecart cart){
        for(EntityMinecart routingCart : routingCarts) {
            if(RailManager.getInstance().areLinked(routingCart, cart)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void func_145843_s(){
        super.func_145843_s();
        if(!field_145850_b.field_72995_K) {
            RailWrapper neighbor = getConnectedRail();
            if(neighbor != null) neighbor.updateSignalCache();
            NetworkController.getInstance(func_145831_w()).updateColor((TileEntitySignalBase)null, func_174877_v());
        }
    }

    @Override
    public void func_73660_a(){
        if(!field_145850_b.field_72995_K) {
            if(firstTick) {
                firstTick = false;
                RailWrapper neighbor = getConnectedRail();
                if(neighbor != null) neighbor.updateSignalCache();
                NetworkController.getInstance(func_145831_w()).updateColor(this, func_174877_v());
            }
            List<EntityMinecart> carts = getNeighborMinecarts();
            for(EntityMinecart cart : carts) {
                if(!routedMinecarts.contains(cart)) {
                    cart.field_71088_bW = 0;
                    onCartEnteringBlock(cart);
                }
            }
            for(EntityMinecart cart : routedMinecarts) {
                if(!carts.contains(cart)) onCartLeavingBlock(cart);
            }
            routedMinecarts = carts;

            RailWrapper neighbor = getConnectedRail();
            if(neighbor != null) {
                updateConnectedSignals(neighbor);
                updateClaims();
            }
        }
    }

    protected EnumLampStatus getLampStatusBlockSignal(List<EntityMinecart> cartsOnNextBlock){
        boolean cartOnNextBlock = !cartsOnNextBlock.isEmpty();

        //If there is a cart on the next block, check if it happens to be part of the same train. If so, still allow the cart to go through.
        if(cartOnNextBlock) {
            List<EntityMinecart> routingCarts = getNeighborMinecarts();
            cartOnNextBlock = !areLinkedCartsPastTheSignal(routingCarts, cartsOnNextBlock);
        }

        return cartOnNextBlock ? EnumLampStatus.RED : EnumLampStatus.GREEN;
    }

    protected abstract void onCartEnteringBlock(EntityMinecart cart);

    protected void onCartLeavingBlock(EntityMinecart cart){
        if(forceMode == EnumForceMode.FORCED_GREEN_ONCE) {
            setForceMode(EnumForceMode.NONE);
        }
        getNextSignals().forEach(signal -> signal.setClaimingCart(null));
    }

    /**
     * Connected in terms of pathfinding
     * @param curRail
     */
    public void updateConnectedSignals(RailWrapper curRail){
        EnumFacing direction = getFacing();
        Set<RailWrapper> rails = new HashSet<>();
        Queue<Map.Entry<RailWrapper, EnumFacing>> traversingRails = new LinkedList<>();
        Set<TileEntitySignalBase> signals = new HashSet<>();
        Map<RailWrapper, SignalBlockNode> blockNodes = new HashMap<>();

        rootSignalNode = new SignalBlockNode(curRail);
        blockNodes.put(curRail, rootSignalNode);

        for(Map.Entry<RailWrapper, EnumFacing> entry : curRail.getNeighbors().entrySet()) {
            if(entry.getValue() != direction.func_176734_d()) {
                traversingRails.add(entry);

                SignalBlockNode neighborSignalBlockNode = new SignalBlockNode(entry.getKey());
                rootSignalNode.nextNeighbors.add(neighborSignalBlockNode);
                blockNodes.put(entry.getKey(), neighborSignalBlockNode);
            }
        }
        rails.add(curRail); //Make sure to consider this block as traversed already, prevents traversing the tracks in reverse direction

        while(!traversingRails.isEmpty()) {
            Map.Entry<RailWrapper, EnumFacing> neighbor = traversingRails.poll();

            TileEntitySignalBase neighborSignal = getNeighborSignal(neighbor.getKey(), neighbor.getValue()); //Find a signal opposing this signal (for like merging splits)
            if(neighborSignal == null) {
                rails.add(neighbor.getKey());
                Map<EnumFacing, TileEntitySignalBase> neighborSignals = neighbor.getKey().getSignals();
                if(neighborSignals.isEmpty()) {
                    for(Map.Entry<RailWrapper, EnumFacing> entry : neighbor.getKey().getNeighborsForEntryDir(neighbor.getValue()).entrySet()) {
                        BlockPos nextNeighbor = entry.getKey();
                        if(!rails.contains(nextNeighbor)) {
                            traversingRails.add(entry);

                            SignalBlockNode neighborSignalBlockNode = new SignalBlockNode(entry.getKey());
                            blockNodes.get(neighbor.getKey()).nextNeighbors.add(neighborSignalBlockNode);
                            blockNodes.put(entry.getKey(), neighborSignalBlockNode);
                        }
                    }
                } else {
                    signals.add(neighborSignals.values().iterator().next());
                }
            }
        }

        if(!signals.equals(nextSignals)) {
            nextSignals = signals;
            sendUpdatePacket();
        }
    }

    /**
     * The signals that are not opposing the direction of this signal.
     * @return
     */
    public Set<TileEntitySignalBase> getNextSignals(){
        return nextSignals;
    }

    @Override
    public SPacketUpdateTileEntity func_189518_D_(){
        NBTTagCompound tag = new NBTTagCompound();
        NBTTagList list = new NBTTagList();

        for(TileEntitySignalBase signal : nextSignals) {
            NBTTagCompound t = new NBTTagCompound();
            t.func_74768_a("x", signal.func_174877_v().func_177958_n());
            t.func_74768_a("y", signal.func_174877_v().func_177956_o());
            t.func_74768_a("z", signal.func_174877_v().func_177952_p());
            list.func_74742_a(t);
        }

        tag.func_74782_a("signals", list);
        rootSignalNode.toNBTTagCompound(tag);
        tag.func_74778_a("text", text);
        tag.func_74778_a("arguments", arguments);
        return new SPacketUpdateTileEntity(func_174877_v(), 0, tag);
    }

    @Override
    public void onDataPacket(NetworkManager net, SPacketUpdateTileEntity pkt){
        if(pkt.func_148853_f() == 0) {
            nextSignals.clear();
            NBTTagList list = pkt.func_148857_g().func_150295_c("signals", 10);
            for(int i = 0; i < list.func_74745_c(); i++) {
                NBTTagCompound t = list.func_150305_b(i);
                BlockPos pos = new BlockPos(t.func_74762_e("x"), t.func_74762_e("y"), t.func_74762_e("z"));
                TileEntity te = field_145850_b.func_175625_s(pos);
                if(te instanceof TileEntitySignalBase) nextSignals.add((TileEntitySignalBase)te);
            }
            rootSignalNode = new SignalBlockNode(pkt.func_148857_g());

            text = pkt.func_148857_g().func_74779_i("text");
            arguments = pkt.func_148857_g().func_74779_i("arguments");
        }
    }

    protected void setMessage(String message, Object... arguments){
        text = message;
        this.arguments = "";
        for(int i = 0; i < arguments.length; i++) {
            if(i > 0) this.arguments += "\n";
            this.arguments += arguments[i].toString();
        }
        field_145850_b.func_184138_a(func_174877_v(), getBlockState(), getBlockState(), 3);
    }

    public String getMessage(){
        String[] localizedArguments = arguments.split("\n");
        for(int i = 0; i < localizedArguments.length; i++) {
            localizedArguments[i] = I18n.func_135052_a(localizedArguments[i]);
        }
        return I18n.func_135052_a(text, (Object[])localizedArguments);
    }

    public SignalBlockNode getSignalBlockInfo(){
        return rootSignalNode;
    }

    @Override
    public void setForceMode(EnumForceMode forceMode){
        this.forceMode = forceMode;
        func_70296_d();
        if(forceMode == EnumForceMode.FORCED_GREEN_ONCE) {
            setLampStatus(EnumLampStatus.GREEN);
            setMessage("signals.signal_message.forced_green");
        } else if(forceMode == EnumForceMode.FORCED_RED) {
            setLampStatus(EnumLampStatus.RED);
            setMessage("signals.signal_message.forced_red");
        } else {
            setMessage("");
        }
    }

    @Override
    public EnumForceMode getForceMode(){
        return forceMode;
    }

    @Override
    public NBTTagCompound func_189515_b(NBTTagCompound tag){
        super.func_189515_b(tag);
        tag.func_74774_a("forceMode", (byte)forceMode.ordinal());

        EntityMinecart cart = getClaimingCart();
        if(cart != null) {
            tag.func_74772_a("ClaimingCartIDMSB", cart.getPersistentID().getMostSignificantBits());
            tag.func_74772_a("ClaimingCartIDLSB", cart.getPersistentID().getLeastSignificantBits());
        }
        return tag;
    }

    @Override
    public void func_145839_a(NBTTagCompound tag){
        super.func_145839_a(tag);
        forceMode = EnumForceMode.values()[tag.func_74771_c("forceMode")];

        if(tag.func_74764_b("ClaimingCartIDMSB")) {
            claimingCartUUID = new UUID(tag.func_74763_f("ClaimingCartIDMSB"), tag.func_74763_f("ClaimingCartIDLSB"));
        }
    }

    protected void setClaimingCart(EntityMinecart cart){
        claimingCart = cart;
    }

    protected EntityMinecart getClaimingCart(){
        if(claimingCartUUID != null) {
            List<EntityMinecart> carts = func_145831_w().func_175644_a(EntityMinecart.class, cart -> cart.func_110124_au().equals(claimingCartUUID));
            claimingCart = carts.isEmpty() ? null : carts.get(0);
            claimingCartUUID = null;
        }

        if(claimingCart != null && claimingCart.field_70128_L) claimingCart = null;
        return claimingCart;
    }

    /**
     * Makes sure that signals don't get stuck getting claimed by carts that will not use this claim.
     * This should mostly be handled by {@link TileEntitySignalBase#onCartLeavingBlock(EntityMinecart)} , but will sometimes not be enough.
     */
    private void updateClaims(){
        if(func_145831_w().func_82737_E() % 100 == 0) {
            EntityMinecart cart = getClaimingCart();
            if(cart != null) {
                AStarRailNode route = RailManager.getInstance().getPath(cart);
                if(route == null) {
                    setClaimingCart(null);
                } else {
                    for(SignalOnRoute signalOnRoute : route.getSignalsOnRoute()) {
                        if(!signalOnRoute.opposite) {
                            if(signalOnRoute.signal.getNextSignals().contains(this)) return;
                        }
                    }

                    setClaimingCart(null);
                }
            }
        }
    }

}
