package com.minemaarten.signals.tileentity;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.ITickable;

import org.apache.commons.lang3.Validate;

import com.minemaarten.signals.api.access.IStationMarker;
import com.minemaarten.signals.capabilities.CapabilityDestinationProvider;
import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.network.GuiSynced;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketSpawnParticle;
import com.minemaarten.signals.rail.RailCacheManager;
import com.minemaarten.signals.rail.RailWrapper;

public class TileEntityStationMarker extends TileEntityBase implements ITickable, IGUITextFieldSensitive,
        IStationMarker{
    private static int nextId;
    @GuiSynced
    private String stationName = "";
    private boolean firstTick = true;

    public TileEntityStationMarker(){
        stationName = "Station" + nextId++;
    }

    @Override
    public NBTTagCompound func_189515_b(NBTTagCompound compound){
        super.func_189515_b(compound);
        compound.func_74778_a("stationName", stationName);
        return compound;
    }

    @Override
    public void func_145839_a(NBTTagCompound compound){
        super.func_145839_a(compound);
        stationName = compound.func_74779_i("stationName");
    }

    @Override
    public String getStationName(){
        return stationName;
    }

    @Override
    public void setStationName(String stationName){
        Validate.notNull(stationName);

        this.stationName = stationName;
        func_70296_d();
        sendUpdatePacket();
    }

    private void updateNeighborRailCache(){
        for(RailWrapper rail : getNeighborRails()) {
            rail.updateStationCache();
        }
    }

    public List<RailWrapper> getNeighborRails(){
        List<RailWrapper> neighbors = new ArrayList<>(1);
        for(EnumFacing d : EnumFacing.values()) {
            RailWrapper rail = RailCacheManager.getInstance(func_145831_w()).getRail(func_145831_w(), func_174877_v().func_177972_a(d));
            if(rail != null) neighbors.add(rail);
        }
        return neighbors;
    }

    @Override
    public void func_145843_s(){
        super.func_145843_s();
        if(!field_145850_b.field_72995_K) {
            updateNeighborRailCache();
            RailCacheManager.getInstance(func_145831_w()).removeStationMarker(this);
        }
    }

    @Override
    public void func_73660_a(){
        if(!field_145850_b.field_72995_K) {
            if(firstTick) {
                firstTick = false;
                updateNeighborRailCache();
                if(!field_145850_b.field_72995_K) {
                    RailCacheManager.getInstance(func_145831_w()).addStationMarker(this);
                }
            }
            updateNeighborMinecarts();
        }
    }

    public void updateNeighborMinecarts(){
        for(EntityMinecart cart : TileEntitySignalBase.getMinecarts(field_145850_b, getNeighborRails())) {
            CapabilityMinecartDestination cap = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null);
            if(cap.getCurrentDestinationRegex().matcher(getStationName()).matches()) {
                cap.nextDestination();
            }
        }
    }

    @Override
    public void setText(int textFieldID, String text){
        setStationName(text);
    }

    @Override
    public String getText(int textFieldID){
        return getStationName();
    }

    @Override
    public NBTTagCompound func_189517_E_(){
        NBTTagCompound tag = super.func_189517_E_();
        tag.func_74778_a("stationName", stationName);
        return tag;
    }

    @Override
    public SPacketUpdateTileEntity func_189518_D_(){
        return new SPacketUpdateTileEntity(func_174877_v(), 0, func_189517_E_());
    }

    @Override
    public void onDataPacket(NetworkManager net, SPacketUpdateTileEntity pkt){
        super.onDataPacket(net, pkt);
        stationName = pkt.func_148857_g().func_74779_i("stationName");
    }

    public boolean isCartApplicable(EntityMinecart cart, Pattern destinationRegex){
        if(destinationRegex.matcher(getStationName()).matches()) return true;
        for(EnumFacing dir : EnumFacing.field_82609_l) {
            TileEntity te = func_145831_w().func_175625_s(func_174877_v().func_177972_a(dir));
            if(te != null) {
                CapabilityDestinationProvider cap = te.getCapability(CapabilityDestinationProvider.INSTANCE, null);
                if(cap != null && cap.isCartApplicable(te, cart, destinationRegex)) {
                    for(int i = 0; i < 10; i++) {
                        double x = func_174877_v().func_177958_n() + func_145831_w().field_73012_v.nextDouble();
                        double z = func_174877_v().func_177952_p() + func_145831_w().field_73012_v.nextDouble();
                        NetworkHandler.sendToAllAround(new PacketSpawnParticle(EnumParticleTypes.ENCHANTMENT_TABLE, x, func_174877_v().func_177956_o() + 1, z, dir.func_82601_c(), dir.func_96559_d(), dir.func_82599_e()), func_145831_w());
                    }
                    return true;
                }
            }
        }
        return false;
    }
}
